
from pathlib import Path
import unittest
import pickle

import numpy as np

from bgen import BgenReader

class TestBgenVar(unittest.TestCase):
    ''' class to make sure BgenVar works correctly
    '''
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        self.folder = Path(__file__).parent /  "data"
      
    def test_minor_allele_dosage(self):
          ''' test we calculate minor_allele_dosage correctly
          '''
          path = self.folder / 'example.16bits.zstd.bgen'
          with BgenReader(path) as bfile:
              for var in bfile:
                  dose = var.minor_allele_dosage
                  probs = var.probabilities
                  
                  # calculate dosages for each allele
                  a1 = (probs[:, 0] * 2 + probs[:, 1])
                  a2 = (probs[:, 2] * 2 + probs[:, 1])
                  
                  # get delta between var.minor_allele_dosage and values calculated here
                  recomputed = a2 if np.nansum(a1) >= np.nansum(a2) else a1
                  delta = abs(dose - recomputed)
                  
                  # check difference between the two estimates is sufficiently low
                  self.assertTrue(np.nanmax(delta) < 2e-7)
    
    def test_alt_dosage(self):
          ''' test we calculate alt_dosage correctly
          '''
          path = self.folder / 'example.16bits.zstd.bgen'
          with BgenReader(path, delay_parsing=True) as bfile:
              for var in bfile:
                  dose = var.alt_dosage
                  probs = var.probabilities
                  
                  # calculate dosages for the alt allele
                  a2 = (probs[:, 2] * 2 + probs[:, 1])
                  
                  # get delta between var.minor_allele_dosage and values calculated here
                  delta = abs(dose - a2)
                  
                  # check difference between the two estimates is sufficiently low
                  self.assertTrue(np.nanmax(delta) < 2.5e-7)
    
    def test_minor_allele_dosage_fast(self):
        ''' test we calculate minor_allele_dosage correctly with the fast path
        '''
        path = self.folder / 'example.8bits.bgen'
        with BgenReader(path) as bfile:
            for var in bfile:
                dose = var.minor_allele_dosage
                probs = var.probabilities
                
                # calculate dosages for each allele
                a1 = (probs[:, 0] * 2 + probs[:, 1])
                a2 = (probs[:, 2] * 2 + probs[:, 1])
                
                # get delta between var.minor_allele_dosage and values calculated here
                recomputed = a2 if np.nansum(a1) >= np.nansum(a2) else a1
                delta = abs(dose - recomputed)
                
                # check difference between the two estimates is sufficiently low
                self.assertTrue(np.nanmax(delta) < 3e-7)
    
    def test_minor_allele_dosage_v11(self):
        ''' test we calculate minor_allele_dosage correctly with version 1 bgens
        '''
        path = self.folder / 'example.v11.bgen'
        with BgenReader(path) as bfile:
            for var in bfile:
                dose = var.minor_allele_dosage
                probs = var.probabilities
                
                # calculate dosages for each allele
                a1 = (probs[:, 0] * 2 + probs[:, 1])
                a2 = (probs[:, 2] * 2 + probs[:, 1])
                
                # get delta between var.minor_allele_dosage and values calculated here
                recomputed = a2 if np.nansum(a1) >= np.nansum(a2) else a1
                delta = abs(dose - recomputed)
                
                # check difference between the two estimates is sufficiently low
                self.assertTrue(np.nanmax(delta) < 7e-5)
    
    def test_pickling(self):
        ''' BgenVar should pickle and unpickle
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            for var in bfile:
                # this checks that we can pickle and unpickle a BgenVar
                pickled = pickle.dumps(var)
                unpickled = pickle.loads(pickled)
                
                # check attributes of the original and unpickled are identical
                self.assertEqual(var.varid, unpickled.varid)
                self.assertEqual(var.rsid, unpickled.rsid)
                self.assertEqual(var.chrom, unpickled.chrom)
                self.assertEqual(var.pos, unpickled.pos)
                self.assertEqual(var.alleles, unpickled.alleles)
                



# cython: language_level=3, boundscheck=False, emit_linenums=True

import logging
from pathlib import Path
import sys
import warnings

from libcpp cimport bool
from libcpp.memory cimport shared_ptr, make_shared
from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint8_t, uint32_t, uint64_t, uintptr_t
from libc.string cimport memcpy

from cython.operator cimport dereference as deref

import numpy as np

from bgen.index import Index

cdef extern from "<iostream>" namespace "std::ios_base":
    cdef cppclass open_mode:
        pass
    cdef open_mode binary
    cdef cppclass iostate:
        pass
    cdef iostate failbit

cdef extern from "<iostream>" namespace "std":
    cdef cppclass istream:
        istream() except +
        void setstate(iostate state) except +

cdef extern from 'variant.h' namespace 'bgen':
    cdef cppclass Variant:
        # declare class constructor and methods
        Variant(istream * handle, uint64_t & offset, int layout, int compression, int expected_n, bool is_stdin) except +
        Variant() except +
        void minor_allele_dosage(float * dosage) except +
        void alt_dosage(float * dosage) except +
        void probs_1d(float * dosage) except +
        int probs_per_sample() except +
        bool phased() except +
        uint8_t * ploidy() except +
        vector[uint8_t] copy_data() except +
        
        # declare public attributes
        string varid, rsid, chrom, minor_allele
        int pos
        long offset
        uint64_t next_variant_offset
        vector[string] alleles
        istream * handle

cdef extern from 'samples.h' namespace 'bgen':
    cdef cppclass Samples:
        Samples(istream * handle, int n_samples) except +
        Samples(string path, int n_samples) except +
        Samples(int n_samples) except +
        Samples() except +
        
        # declare public attributes
        vector[string] samples

cdef extern from 'header.h' namespace 'bgen':
    cdef cppclass Header:
        Header(istream * handle) except +
        Header();
        uint32_t offset
        uint32_t nvariants
        uint32_t nsamples
        int compression
        int layout
        string extra
        bool has_sample_ids

cdef extern from 'reader.h' namespace 'bgen':
    cdef cppclass CppBgenReader:
        # declare class constructor and methods
        CppBgenReader(string path, string sample_path, bool delay_parsing) except +
        void parse_all_variants()
        Variant & operator[](int idx)
        Variant & get(int idx)
        void drop_variants(vector[int] indices)
        vector[string] varids()
        vector[string] rsids()
        vector[string] chroms()
        vector[uint32_t] positions()
        
        # declare public attributes
        istream * handle
        vector[Variant] variants
        Samples samples
        Header header
        uint64_t offset

cdef extern from 'utils.h' namespace 'bgen':
    cdef struct Range:
        uint8_t _min
        uint8_t _max
    uint64_t fast_ploidy_sum(uint8_t * ploidy, uint32_t & size) except +
    Range fast_range(uint8_t * ploidy, uint32_t & size)

cdef class IStream:
    ''' basic cython implementation of std::istream, for easy pickling
    '''
    cdef istream * ptr
    def __cinit__(self, uint64_t ptr):
        self.ptr = <istream*>ptr
    
    def __str__(self):
        return f'std::istream at {<uint64_t>self.ptr}'
    
    def __dealloc__(self):
        pass
    
    def __reduce__(self):
        return (self.__class__, (<uint64_t>self.ptr, ))

cdef class OpenStatus:
    ''' class to share status of whether a bgen file is currently open
    
    This uses a shared pointer so we can store the status in a BgenReader
    object, but also pass that status to BgenVariant objects. When a bgen
    file is closed, the status changes for all loaded variants, in order
    to prevent reading more data from the closed file.
    
    This may not work for pickled BgenVariant objects (after unpickling), as
    the BgenReader object may have been closed after the BgenVariant was 
    pickled. In order to allow pickled BgenVariants, as part of unpickling 
    the BgenVariant, it creates a new OpenStatus object where status=True.
    '''
    cdef shared_ptr[bool] ptr
    def __cinit__(self):
        self.ptr = make_shared[bool](True)
    def __str__(self):
        return f'status={deref(self.ptr)}'
    def __eq__(self, other):
        return deref(self.ptr) == other
    def off(self):
        self.ptr = make_shared[bool](False)
    def __reduce__(self):
        return (self.__class__, ())

cdef class BgenHeader:
    ''' holds information about the Bgen file, obtained from the intial header.
    '''
    cdef uint32_t _offset
    cdef uint32_t _nvariants
    cdef uint32_t _nsamples
    cdef int _compression
    cdef int _layout
    cdef bool _has_sample_ids
    cdef string _metadata
    cdef object compress_formats
    def __cinit__(self, uint32_t offset, uint32_t nvariants, uint32_t nsamples,
            int compression, int layout, bool has_sample_ids, string metadata):
        self._offset = offset
        self._nvariants = nvariants
        self._nsamples = nsamples
        self._compression = compression
        self._layout = layout
        self._has_sample_ids = has_sample_ids
        self._metadata = metadata
        self.compress_formats = {0: None, 1: 'zlib', 2: 'zstd'}
    
    def __repr__(self):
        return f'BgenHeader(offset={self.offset}, nvariants={self.nvariants}, ' \
            f'nsamples={self.nsamples}, compression={self.compression}, ' \
            f'layout={self.layout}, has_sample_ids={self.has_sample_ids})'
    
    @property
    def offset(self): return self._offset
    @property
    def nsamples(self): return self._nsamples
    @property
    def nvariants(self): return self._nvariants
    @property
    def compression(self): return [None, 'zlib', 'zstd'][self._compression]
    @property
    def compression(self): return self.compress_formats[self._compression]
    @property
    def layout(self): return self._layout
    @property
    def has_sample_ids(self): return self._has_sample_ids
    @property
    def metadata(self): return self._metadata.decode('utf8')

cdef class BgenVar:
    ''' holds data for a Variant from a bgen file
    
    This constructs a new Variant, rather than using a object pointer, in order
    to make pickling the object easier.
    
    Initialization takes about 1e-5 seconds per variant, so we can only run
    through variants in a file at 100,000 variants per second at most (assuming
    no other work is being done).
    
    This shouldn't be a limitation in practise, since the rate limiting part is
    parsing genotype information, which runs at about 500 variants per second
    for files with 500,000 samples.
    '''
    cdef Variant * thisptr
    cdef IStream handle
    cdef uint64_t offset
    cdef int layout, compression, expected_n
    cdef bool is_stdin
    cdef OpenStatus is_open
    def __cinit__(self,
                  IStream handle,
                  uint64_t offset,
                  int layout,
                  int compression,
                  int expected_n,
                  bool is_stdin,
                  OpenStatus is_open,
                  ):
        self.handle = handle
        self.offset = offset
        self.layout = layout
        self.compression = compression
        self.expected_n = expected_n
        self.is_stdin = is_stdin
        self.is_open = is_open
        
        # construct new Variant from the handle, offset and other file info
        self.thisptr = new Variant(self.handle.ptr, offset, layout, compression, expected_n, is_stdin)
    
    def __repr__(self):
       return f'BgenVar("{self.varid}", "{self.rsid}", "{self.chrom}", {self.pos}, {self.alleles})'
    
    def __str__(self):
       return f'{self.rsid} - {self.chrom}:{self.pos} {self.alleles}'
    
    def __reduce__(self):
        ''' enable pickling of a BgenVar object
        '''
        warnings.warn("pickling BgenVar - make sure their BgenReader objects exist when unpickling", RuntimeWarning)
        return (self.__class__, (self.handle, self.thisptr.offset, self.layout, self.compression, self.expected_n, self.is_stdin, self.is_open))
    
    def __dealloc__(self):
        del self.thisptr
    
    @property
    def varid(self):
      return self.thisptr.varid.decode('utf8')
    @property
    def rsid(self):
        return self.thisptr.rsid.decode('utf8')
    @property
    def chrom(self):
        return self.thisptr.chrom.decode('utf8')
    @property
    def pos(self):
        return self.thisptr.pos
    @property
    def fileoffset(self):
        return self.offset
    @property
    def next_variant_offset(self):
        return self.thisptr.next_variant_offset
    @property
    def alleles(self):
        return [x.decode('utf8') for x in self.thisptr.alleles]
    cdef __check_closed(self):
        ''' make sure we cannot read from a closed bgen file
        
        If the bgen file is closed before we try to access the genotype data,
        then we should not be able to read from the file. However, this
        object holds a pointer to a istream object, and it does not know if
        the underlying memory is valid, so could try to access it anyway.
        
        We prevent this by checking a shared pointer held by all the BgenVars 
        which were opened by a given BgenReader. This shared pointer will have
        been set to false when the bgen file closed, so we can check that
        shared status, and set the istream iostate to fail. Later cpp calls which
        would read from the stale file handle will raise ValuErrors instead due
        to checking the failbit before attempting to read from the istream.
        
        However, this does not work for pickled BgenVars, since we cannot know
        whether the istream object exists or not.
        '''
        if not deref(self.is_open.ptr):
            self.thisptr.handle.setstate(failbit)
    @property
    def is_phased(self):
        return self.thisptr.phased()
    @property
    def ploidy(self):
        ''' get the ploidy for each sample
        '''
        self.__check_closed()
        cdef uint8_t * ploid = self.thisptr.ploidy()
        cdef uint64_t size = self.expected_n
        cdef uint8_t[::1] arr = np.empty(size, dtype=np.uint8, order='C')
        memcpy(&arr[0], ploid, size)
        return np.asarray(arr)
    @property
    def minor_allele(self):
        ''' get the minor allele of a biallelic variant
        '''
        return self.thisptr.minor_allele.decode('utf8')
    @property
    def minor_allele_dosage(self):
        ''' dosage for the minor allele for a biallelic variant
        '''
        self.__check_closed()
        cdef float[:] dose = np.empty(self.expected_n, dtype=np.float32, order='C')
        self.thisptr.minor_allele_dosage(&dose[0])
        return np.asarray(dose)
    @property
    def alt_dosage(self):
        ''' dosage for the alt allele for a biallelic variant
        '''
        self.__check_closed()
        cdef float[:] dose = np.empty(self.expected_n, dtype=np.float32, order='C')
        self.thisptr.alt_dosage(&dose[0])
        return np.asarray(dose)
    @property
    def probabilities(self):
        ''' get the allelic probabilities for a variant
        '''
        self.__check_closed()
        cdef int cols = self.thisptr.probs_per_sample()
        cdef uint32_t n_samples = self.expected_n
        cdef uint64_t size = n_samples * cols
        cdef uint8_t[::1] ploidy
        if self.is_phased:
            ploidy = self.ploidy
            size = fast_ploidy_sum(&ploidy[0], n_samples) * cols
        
        cdef float[:] arr = np.empty(size, dtype=np.float32, order='C')
        self.thisptr.probs_1d(&arr[0])
        
        cdef int current = 0
        cdef int phase_width
        cdef Range ploidy_range
        if self.is_phased:
            ploidy_range = fast_range(&ploidy[0], n_samples)
            max_ploidy = ploidy_range._max
            min_ploidy = ploidy_range._min
            if min_ploidy == max_ploidy:
                # quickly reshape probs if ploidy is constant
                width = max_ploidy * cols
                data = np.reshape(arr, (-1, width))
            else:
                # phased data initially comes as one row per haploytpe. This is
                # reshaped to concatenate haplotype data into single row. Fill
                # in a new array from the old data row by row
                data = np.reshape(arr, (-1, cols))
                phase_width = data.shape[1]
                
                # create an empty array filled with nans
                ragged = np.empty((len(ploidy), max_ploidy * cols))
                ragged.fill(np.nan)
                
                # fill in the empty array
                for i, x in enumerate(ploidy):
                    for y in range(x):
                        start = y * phase_width
                        end = start + phase_width
                        ragged[i, start:end] = data[current]
                        current += 1
                
                data = ragged
        else:
            data = np.reshape(arr, (-1, cols))
        
        return data
    
    def copy_data(self):
        ''' get a copy of the data on disk for the variant.
        
        This can be used to quickly copy data from one bgen into another, if
        you have the same set of samples between the source and destination
        bgens. This primarily avoids decompressing, decoding, re-encoding, and
        compressing the genotype data.
        '''
        self.__check_closed()
        cdef vector[uint8_t] data = self.thisptr.copy_data()
        return data

cdef class BgenReader:
    ''' class to open bgen files from disk, and access variant data within
    '''
    cdef CppBgenReader * thisptr
    cdef string path, sample_path
    cdef bool delay_parsing, is_stdin
    cdef IStream handle
    cdef object index
    cdef OpenStatus is_open
    cdef uint64_t offset
    def __cinit__(self, path, sample_path='', bool delay_parsing=False):
        if isinstance(path, Path):
            path = str(path)
        if isinstance(sample_path, Path):
            sample_path = str(sample_path)
        self.is_stdin = self.__is_from_stdin(path)
        if self.is_stdin:
            delay_parsing = True
            path = '/dev/stdin'
        
        if Path(path).exists() and Path(path).is_dir():
            raise ValueError(f'bgen path is for a folder: {path}')
        
        delay_parsing |= self._check_for_index(path)
        
        self.path = path.encode('utf8')
        self.sample_path = sample_path.encode('utf8')
        self.delay_parsing = delay_parsing
        
        samp = '' if sample_path == '' else f', (samples={self.sample_path.decode("utf")})'
        logging.debug(f'opening BgenFile from {self.path.decode("utf")}{samp}')
        self.thisptr = new CppBgenReader(self.path, self.sample_path, self.delay_parsing)
        self.handle = IStream(<uint64_t>self.thisptr.handle)
        self.is_open = OpenStatus()
        self.offset = self.thisptr.offset
    
    def __is_from_stdin(self, bgen_path):
        if bgen_path is sys.stdin:
            return True
        elif str(bgen_path) == '/dev/stdin':
            return True
        elif str(bgen_path) == '-':
            return True
        return False
    
    def __dealloc__(self):
        self.close()
    
    def __repr__(self):
        return f'BgenFile("{self.path.decode("utf8")}", "{self.sample_path.decode("utf8")}")'
    
    def __iter__(self):
        return self
    
    def __next__(self):
        ''' iterate through all variants in the bgen file
        '''
        if not self.is_open == True:
            raise ValueError('bgen file is closed')
        
        try:
            var = BgenVar(self.handle, self.offset, self.thisptr.header.layout,
                self.thisptr.header.compression, self.thisptr.header.nsamples, self.is_stdin,
                self.is_open)
            self.offset = var.next_variant_offset
            return var
        except IndexError:
            raise StopIteration
    
    def __len__(self):
      if not self.is_open == True:
          raise ValueError("bgen file is closed")
      
      length = self.thisptr.variants.size()
      if length > 0:
          return length
      else:
          return self.thisptr.header.nvariants
    
    def __getitem__(self, int idx):
        ''' pull out a Variant by index position
        '''
        if not self.is_open == True:
            raise ValueError('bgen file is closed')
        
        if idx >= len(self) or idx < 0:
            raise IndexError(f'cannot get Variant at index: {idx}')
        
        # account for lazy loading variants from bgen
        if self.index is None and self.thisptr.variants.size() == 0:
            self.thisptr.parse_all_variants()
        
        cdef long offset
        offset = self.index.offset_by_index(idx) if self.index else self.thisptr.variants[idx].offset
        return BgenVar(self.handle, offset, self.thisptr.header.layout,
          self.thisptr.header.compression, self.thisptr.header.nsamples,
          self.is_stdin, self.is_open)
    
    def _check_for_index(self, bgen_path):
        ''' creates self.index if a bgenix index file is available
        '''
        index_path = Path(bgen_path + '.bgi')
        idx_exists = index_path.exists()
        self.index = Index(index_path) if idx_exists else None
        return idx_exists
    
    @property
    def header(self):
      ''' get header info from bgen file
      '''
      if not self.is_open == True:
          raise ValueError("bgen file is closed")
      
      hdr = self.thisptr.header
      return BgenHeader(hdr.offset, hdr.nvariants, hdr.nsamples,
          hdr.compression, hdr.layout, hdr.has_sample_ids, hdr.extra)
    
    @property
    def samples(self):
      ''' get list of samples in the bgen file
      '''
      if not self.is_open == True:
          raise ValueError("bgen file is closed")
      
      samples = self.thisptr.samples.samples
      return [x.decode('utf8') for x in samples]
    
    def drop_variants(self, list indices):
        ''' drops variants from bgen by indices, for a a =ing processing variants
        '''
        if not self.is_open == True:
            raise ValueError("bgen file is closed")
        
        if self.delay_parsing:
            self.thisptr.parse_all_variants()
        
        self.thisptr.drop_variants(indices)
    
    def fetch(self, chrom, start=None, stop=None):
        ''' fetches all variants within a genomic region
        
        Args:
            chrom: chromosome that variants must be on
            start: start nucleotide of region. If None, gets all variants on chromosome
            stop: end nucleotide of region. If None, gets variants with positions after start
        
        Yields:
            BgenVars for variants within the genome region
        '''
        if not self.is_open == True:
            raise ValueError('bgen file is closed')
        
        if not self.index:
            raise ValueError("can't fetch variants without index")
        
        for offset in self.index.fetch(chrom, start, stop):
            yield BgenVar(self.handle, offset, self.thisptr.header.layout,
                self.thisptr.header.compression, self.thisptr.header.nsamples,
                self.is_stdin, self.is_open)
    
    def with_rsid(self, rsid):
      ''' get BgenVar from file given an rsID
      '''
      if not self.is_open == True:
          raise ValueError('bgen file is closed')
      
      if self.index:
          offsets = self.index.offset_by_rsid(rsid)
          return [BgenVar(self.handle, int(offset), self.thisptr.header.layout,
                          self.thisptr.header.compression, self.thisptr.header.nsamples,
                          self.is_stdin, self.is_open)
                  for offset in offsets]
      
      if not self.delay_parsing:
          idx = [i for i, x in enumerate(self.rsids()) if x == rsid]
          return [self[i] for i in idx]
      
      raise ValueError("can't get variant without fully loading the bgen, or indexing")
    
    def at_position(self, pos):
      ''' get BgenVar from file given a position
      '''
      if not self.is_open == True:
          raise ValueError('bgen file is closed')
      
      if self.index:
          offsets = self.index.offset_by_pos(pos)
          return [BgenVar(self.handle, int(offset), self.thisptr.header.layout,
                          self.thisptr.header.compression, self.thisptr.header.nsamples,
                          self.is_stdin, self.is_open) 
                  for offset in offsets]
      
      if not self.delay_parsing:
          idx = [i for i, x in enumerate(self.positions) if x == pos]
          return [self[i] for i in idx]
      
      raise ValueError("can't get variant without fully loading the bgen, or indexing")
    
    def varids(self):
      ''' get the variant IDs of all variants in the bgen file
      '''
      if not self.is_open == True:
          raise ValueError("bgen file is closed")
      
      if self.index:
          raise ValueError("can't load varids when using an index file")
      
      varids = self.thisptr.varids()
      return [x.decode('utf8') for x in varids]
    
    def rsids(self):
      ''' get the rsIDs of all variants in the bgen file
      '''
      if not self.is_open == True:
          raise ValueError("bgen file is closed")
      
      if self.index:
          return self.index.rsids
      
      rsids = self.thisptr.rsids()
      return [x.decode('utf8') for x in rsids]
    
    def chroms(self):
        ''' get the chromosomes of all variants in the bgen file
        '''
        if not self.is_open == True:
            raise ValueError("bgen file is closed")
        
        if self.index:
            return self.index.chroms
        
        chroms = self.thisptr.chroms()
        return [x.decode('utf8') for x in chroms]
    
    def positions(self):
        ''' get the positions of all variants in the bgen file
        '''
        if not self.is_open == True:
            raise ValueError("bgen file is closed")
        
        if self.index:
            return self.index.positions
        
        return self.thisptr.positions()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False
    
    def close(self):
        if self.is_open == True:
            del self.thisptr
            self.handle = None
            if self.index:
                self.index.close()
            self.index = None
            self.is_open.off()

BgenFile = BgenReader




from pathlib import Path
import unittest

import numpy as np

from bgen import BgenReader

def get_alt_dose(variant):
    ''' compute alt dosage from variant probabilities
    
    This accounts for ploidy and phased status.
    '''
    probs = variant.probabilities
    haploid = variant.ploidy == 1
    
    alt_dose = np.empty(len(probs))
    if variant.is_phased:
        alt_dose[~haploid] = probs[~haploid, 1] + probs[~haploid, 3]
        alt_dose[haploid] = probs[haploid, 1]
    else:
        alt_dose[~haploid] = 2 * probs[~haploid, 2] + probs[~haploid, 1]
        alt_dose[haploid] = probs[haploid, 1]
    
    return alt_dose

class TestAltDosage(unittest.TestCase):
    ''' class to make sure alt dosage is computed correctly
    '''
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        self.folder = Path(__file__).parent /  "data"
    
    def test_alt_dosage_nonstandard(self):
        ''' variant.alt_dosage is correct with variable ploidy and with phased data
        '''
        path = self.folder / 'alt_dosage_check.bgen'
        with BgenReader(path) as bfile:
            for variant in bfile:
                dose = variant.alt_dosage
                self.assertTrue((dose >= 0).all())
                self.assertTrue((dose == get_alt_dose(variant)).all())




from pathlib import Path
import os
import unittest
import sys

from bgen import BgenReader

from tests.utils import load_gen_data, arrays_equal

class TestBgenStream(unittest.TestCase):
    ''' class to make sure BgenReader works correctly
    '''
    
    @classmethod
    def setUpClass(cls):
        cls.gen_data = load_gen_data()
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        # can't write > 2 ** 16 bytes to linux write_fd pipe before something stalls
        self.max_buff = 65536
        self.folder = Path(__file__).parent /  "data"
    
    @unittest.skipIf(sys.platform == "win32", "haven't figured out file handle " \
                                              "duplication and writing on windows " \
                                              "at the required buffer size")
    def test_bgen_streaming(self):
        ''' check that was can open a bgen file from stdin
        '''
        # Save original stdin and stdout file descriptors
        original_stdin_fd = os.dup(sys.stdin.fileno())
        original_stdout_fd = os.dup(sys.stdout.fileno())
        read_fd, write_fd = os.pipe()
        
        path = self.folder / 'example.16bits.zstd.bgen'
        with open(path, 'rb') as handle:
            data = handle.read()
        
        try:
            os.write(write_fd, data[:self.max_buff])
            os.close(write_fd) # Close the write end once done writing
            
            # Redirect stdin file descriptor to the read end of the pipe
            os.dup2(read_fd, sys.stdin.fileno())
            os.close(read_fd) # Close the original read end after duplicating
            
            with BgenReader(sys.stdin) as bfile:
                header = bfile.header
                self.assertEqual(header.nvariants, 199)
                self.assertEqual(header.nsamples, 500)
                
                var1 = next(bfile)
                var1.alt_dosage # load dosage for the first variant
                g = self.gen_data[0]
                self.assertTrue(arrays_equal(g.probabilities, var1.probabilities, bit_depth=16))
                
                var2 = next(bfile)  # don't load any genotype probabilities for this variant
        finally:
            # Restore original stdin and stdout file descriptors
            os.dup2(original_stdin_fd, sys.stdin.fileno())
            os.close(original_stdin_fd)
            sys.stdout = sys.__stdout__ # Use sys.__stdout__ to get the true original stdout
            # Close the duplicated original file descriptors
            os.close(original_stdout_fd)
        
        # can't get header for closed bgen file
        with self.assertRaises(ValueError):
            bfile.header
        self.assertEqual(var1.pos, 2000)
        self.assertEqual(var2.pos, 3000)
        
        # we should still be able to access genotype probabilities for the first variant, as
        # these were loaded while the file was still open
        g = self.gen_data[0]
        self.assertTrue(arrays_equal(g.probabilities, var1.probabilities, bit_depth=16))
        
        # # in contrast, we shouldn't be able to load the genotype probabilities for the 
        # # second variant, as these were not loaded while the bgen was open. It should raise
        # # and error, but currently does not. I haven't figured out why not yet.
        # with self.assertRaises(ValueError):
        #     var2.probabilities
        



from pathlib import Path
import unittest

import numpy as np

from bgen import BgenReader
from bgen.index import Index

from tests.utils import load_gen_data

class TestBgenIndex(unittest.TestCase):
    ''' class to make sure bgen.index.Index works correctly
    '''
    @classmethod
    def setUpClass(cls):
        cls.gen_data = load_gen_data()
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        self.folder = Path(__file__).parent /  "data"
    
    def test_index_opens(self):
        ''' loads index when available
        '''
        bfile = BgenReader(self.folder / 'example.15bits.bgen')
        self.assertFalse(bfile._check_for_index(str(self.folder / 'example.15bits.bgen')))
        
        bfile = BgenReader(self.folder / 'example.16bits.bgen')
        self.assertTrue(bfile._check_for_index(str(self.folder / 'example.16bits.bgen')))
    
    def test_index_fetch(self):
        ''' fetches file offsets
        '''
        chrom = '01'
        start = 5000
        stop = 50000
        
        index = Index(self.folder / 'example.16bits.bgen.bgi')
        self.assertTrue(len(list(index.fetch(chrom))) == len(self.gen_data))
        self.assertTrue(len(list(index.fetch('02'))) == 0)
        self.assertTrue(len(list(index.fetch(chrom, start * 100, stop * 100))) == 0)
        
        # check for a whole chromosome
        chrom_offsets = list(index.fetch(chrom))
        self.assertTrue(len(chrom_offsets) > 0)
        self.assertTrue(len(chrom_offsets) == len(self.gen_data))
        
        # check for all variants following a position
        after_pos_offsets = list(index.fetch(chrom, start))
        self.assertTrue(len(after_pos_offsets) > 0)
        self.assertTrue(len(after_pos_offsets) == len([x for x in self.gen_data if start <= x.pos]))
        
        # check for all variants within a region
        in_region_offsets = list(index.fetch(chrom, start, stop))
        self.assertTrue(len(in_region_offsets) > 0)
        self.assertTrue(len(in_region_offsets) == len([x for x in self.gen_data if start <= x.pos <= stop]))
        
        # make sure the queries return different lists
        self.assertTrue(len(chrom_offsets) != len(after_pos_offsets))
        self.assertTrue(len(chrom_offsets) != len(in_region_offsets))
        self.assertTrue(len(after_pos_offsets) != len(in_region_offsets))
        




from pathlib import Path
import unittest

import numpy as np

from bgen import BgenReader

from tests.utils import load_gen_data, load_vcf_data, load_haps_data, arrays_equal

class TestExampleBgens(unittest.TestCase):
    ''' class to make sure we can load bgen files
    '''
    @classmethod
    def setUpClass(cls):
        cls.gen_data = load_gen_data()
        cls.vcf_data = load_vcf_data()
        cls.haps_data = load_haps_data()
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        self.folder = Path(__file__).parent /  "data"
    
    def test_load_example_genotypes_bit_depths(self):
        ''' check parsing genotypes from the example files with different bit depths
        '''
        for path in self.folder.glob('example.*bits.bgen'):
            bit_depth = int(path.stem.split('.')[1].strip('bits'))
            bfile = BgenReader(str(path))
            for var, g in zip(bfile, self.gen_data):
                self.assertEqual(g, var)
                self.assertTrue(arrays_equal(g.probabilities, var.probabilities, bit_depth))
    
    def test_zstd_compressed(self):
        ''' check we can parse genotypes from zstd compressed geno probabilities
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        bfile = BgenReader(str(path))
        for var, g in zip(bfile, self.gen_data):
            self.assertEqual(g, var)
            self.assertTrue(arrays_equal(g.probabilities, var.probabilities, 16))
    
    def test_v11(self):
        ''' check we can open a bgen in v1.1 format, and parse genotypes correctly
        '''
        path = self.folder / 'example.v11.bgen'
        bfile = BgenReader(str(path))
        bit_depth = 16
        for var, g in zip(bfile, self.gen_data):
            self.assertEqual(g, var)
            self.assertTrue(arrays_equal(g.probabilities, var.probabilities, bit_depth))
    
    def test_Path(self):
        ''' check we can open bgen files from Path objects
        '''
        path = self.folder / 'example.v11.bgen'
        bfile = BgenReader(path)
    
    def test_load_haplotypes_bgen(self):
        ''' check we can open a bgen with haplotypes, and parse genotypes correctly
        '''
        path = self.folder / 'haplotypes.bgen'
        bfile = BgenReader(str(path))
        bit_depth = 16
        for var, g in zip(bfile, self.haps_data):
            self.assertEqual(g, var)
            self.assertTrue(arrays_equal(g.probabilities, var.probabilities, bit_depth))
    
    def test_load_complex_file(self):
        ''' make sure we can open a complex bgen file
        '''
        path = self.folder / 'complex.bgen'
        bfile = BgenReader(path)
        bit_depth = 16
        for var, g in zip(bfile, self.vcf_data):
            self.assertEqual(g, var)
            self.assertTrue(arrays_equal(g.probabilities, var.probabilities, bit_depth))
            self.assertTrue(all(x == y for x, y in zip(g.ploidy, var.ploidy)))
    
    def test_load_complex_files(self):
        ''' make sure we can open the complex bgen files
        '''
        
        for path in self.folder.glob('complex.*.bgen'):
            bit_depth = int(path.stem.split('.')[1].strip('bits'))
            bfile = BgenReader(path)
            for var, g in zip(bfile, self.vcf_data):
                self.assertEqual(g, var)
                self.assertTrue(arrays_equal(g.probabilities, var.probabilities, bit_depth))
    
    def test_load_missing_file(self):
        ''' check passing in a path to a missing file fails gracefully
        '''
        with self.assertRaises(ValueError):
            BgenReader('/zzz/jjj/qqq.bgen')
    
    # def test_load_missing_sample_file(self):
    #     path = str(self.folder / 'example.8bits.bgen')
    #     bfile = BgenReader(path, '/zzz/jjj/qqq.sample')




import bisect
import sqlite3
import logging

import numpy as np

class Index:
    def __init__(self, path):
        logging.debug(f'opening bgen index: {path}')
        self.path = str(path)
        self.conn = sqlite3.connect(self.path)
        self.cur = self.conn.cursor()
        self.dropped_variants = None
        
        self._rsids = None
        self._chroms = None
        self._positions = None
        self._masked = None
    
    def fetch(self, chrom, start=None, stop=None):
        ''' get file offsets for variants within a genome region in a bgen file
        
        Args:
            chrom: chromosome that variants must be on
            start: start nucleotide of region. If None, gets offsets for all
                variants on chromosome
            stop: end nucleotide of region. If None, gets offsets for variants
                with positions after start
        
        Yields:
            file offsets for variants within the genome region
        '''
        if start is None and stop is None:
            query = 'SELECT file_start_position FROM Variant WHERE chromosome=?'
            params = (chrom, )
        elif stop is None:
            query = 'SELECT file_start_position FROM Variant \
                        WHERE chromosome=? AND position >= ?'
            params = (chrom, start)
        else:
            query = 'SELECT file_start_position FROM Variant \
                        WHERE chromosome=? AND position >= ? AND position <= ?'
            params = (chrom, start, stop)

        query = self.cur.execute(query, params)
        while True:
            res = query.fetchone()
            if not res:
                break
            yield res[0]
    
    def offset_by_index(self, index):
        ''' get file offset of bgen variant given a variant index
        '''
        offset = self.cur.execute('''SELECT file_start_position FROM Variant LIMIT 1 OFFSET ?''', (index, )).fetchone()
        return offset[0]
        
    def offset_by_rsid(self, rsid):
        ''' get file offset of bgen variant given a variant index
        '''
        offsets = self.cur.execute("SELECT file_start_position FROM Variant WHERE rsid= ?", (rsid, )).fetchall()
        return [x[0] for x in offsets]
    
    def offset_by_pos(self, pos):
        ''' get file offset of bgen variant given a variant index
        '''
        offsets = self.cur.execute("SELECT file_start_position FROM Variant WHERE position= ?", (pos, )).fetchall()
        return [x[0] for x in offsets]
    
    @property
    def rsids(self):
        ''' get rsID list for all variants in the bgen file
        '''
        if self._rsids is None:
            query = self.cur.execute(
                "SELECT rsid FROM Variant ORDER BY file_start_position")
            self._rsids = [x[0] for x in query.fetchall()]
        return self._rsids
    
    @property
    def chroms(self):
        ''' get chromosome list for all variants in the bgen file
        '''
        if self._chroms is None:
            query = self.cur.execute(
                "SELECT chromosome FROM Variant ORDER BY file_start_position")
            self._chroms = [x[0] for x in query.fetchall()]
        return self._chroms
    
    @property
    def positions(self):
        ''' get position list for all variants in the bgen file
        '''
        if self._positions is None:
            query = self.cur.execute(
                "SELECT position FROM Variant ORDER BY file_start_position")
            self._positions = np.array([x[0] for x in query.fetchall()])
        return self._positions

    def close(self):
        if self.cur is not None:
            self.cur.close()
        self.cur = None
        if self.conn is not None:
            self.conn.close()
        self.conn = None





from pathlib import Path
import unittest
import tempfile
import os
import time
import sys
import math

import numpy as np

from bgen import BgenReader, BgenWriter
from bgen.index import Index

def probs_close(orig, updat, bit_depth):
    ''' check if the genotype probabilities are near the original values

    We lose some precision when storing genotype probabilities, dependent on the
    bit depth we select. This accounts for the variable bit depth.
    '''
    # fix the nans from differing ploidies
    orig = orig.copy()
    orig[np.isnan(orig)] = 1
    updat[np.isnan(updat)] = 1

    # check abolute error (accounting for bit depth)
    max_error = 1 / ((2 ** bit_depth) - 1)
    matched = (abs(orig - updat) < max_error)

    # get relative difference between original and stored values
    delta = abs(updat / orig)
    delta[delta < 1] = 1 / delta[delta < 1]
    delta -= 1
    delta[np.isinf(delta)] = 0

    int_vals = np.floor((2 ** bit_depth - 1) * orig)
    max_delta = 1.0001 / int_vals

    # allow it through if the abolute error is sufficiently low, or the relative
    # error is sufficiently low, of if they differ by less than 1 part in 10 million
    matched = matched | (delta <= max_delta) | (delta < 1e-7)
    return matched.all()

class TestBgenWriter(unittest.TestCase):
    ''' class to make sure bgen.BgenWriter works correctly
    '''

    def setUp(self):
        self.tmp = tempfile.TemporaryDirectory()
        self.tmpdir = Path(self.tmp.name)
    
    def tearDown(self):
        try:
            self.tmp.cleanup()
        except:
            pass
    
    def test_writing(self):
        ''' test basic BgenWriter file without variants
        '''
        path = self.tmpdir / 'temp.bgen'

        with BgenWriter(path, n_samples=3, samples=['a', 'b', 'c']) as bfile:
            pass
        
        with BgenReader(path, delay_parsing=True) as bfile:
            # check all the header attributes look ok
            self.assertEqual(bfile.samples, ['a', 'b', 'c'])
            self.assertEqual(bfile.header.offset, 37)
            self.assertEqual(bfile.header.nsamples, 3)
            self.assertEqual(bfile.header.nvariants, 0)
            self.assertEqual(bfile.header.compression, 'zstd')
            self.assertEqual(bfile.header.layout, 2)
            self.assertTrue(bfile.header.has_sample_ids)
            self.assertEqual(bfile.header.metadata, '')

        path2 = self.tmpdir / 'temp2.bgen'

        # check if we change the attributes, then we get the right data
        # bfile = BgenWriter(tmp.name, n_samples=4, samples=['a', 'b', 'c', 'd'], 
        with BgenWriter(path2, n_samples=4, samples=[], 
                           compression=None, layout=1, metadata='1234') as bfile:
            pass
        
        with BgenReader(path2, delay_parsing=True) as bfile:
            # check the new header attributes look ok
            self.assertEqual(bfile.samples, ['0', '1', '2', '3'])
            self.assertEqual(bfile.header.offset, 24)
            self.assertEqual(bfile.header.nsamples, 4)
            self.assertEqual(bfile.header.nvariants, 0)
            self.assertIsNone(bfile.header.compression)
            self.assertEqual(bfile.header.layout, 1)
            self.assertFalse(bfile.header.has_sample_ids)
            self.assertEqual(bfile.header.metadata, '1234')

    def test_writing_to_closed_file(self):
        ''' check we can't write to a closed file
        '''
        path = self.tmpdir / 'temp.bgen'
        sample_ids = ['a', 'b', 'c']
        with BgenWriter(path, n_samples=3, samples=sample_ids) as bfile:
            geno = np.array([[0.1, 0.8, 0.1], [0.5, 0.25, 0.25], [0.1, 0.2, 0.7]])
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,)

        with self.assertRaises(ValueError):
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,)
    
    def test_writing_metadata_with_newline(self):
        ''' check we can write metadata (and read back!)
        '''
        sample_ids = ['a', 'b', 'c']
        metadata = 'a\nbc'  # previously had errors with metadata containing newlines
        path = self.tmpdir / 'temp.bgen'
        
        bit_depth = 16
        geno = np.array([[0.1, 0.8, 0.1], [0.5, 0.25, 0.25], [0.1, 0.2, 0.7]])
        with BgenWriter(path, n_samples=3, samples=sample_ids, metadata=metadata) as bfile:
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno, bit_depth=bit_depth)
        
        bfile = BgenReader(path)
        self.assertEqual(metadata, bfile.header.metadata)
        var = next(bfile)
        self.assertTrue(probs_close(geno[:, :-1], var.probabilities[:, :-1], bit_depth))
    
    def test_wrong_sample_number(self):
        ''' check we can't write variants with the wrong number of samples
        '''
        path = self.tmpdir / 'temp.bgen'
        sample_ids = ['a', 'b', 'c']
        with BgenWriter(path, n_samples=3, samples=sample_ids) as bfile:
            geno = np.array([[0.1, 0.8, 0.1], [0.5, 0.25, 0.25]])

            with self.assertRaises(ValueError):
                bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,)

    def test_writing_variant_attributes(self):
        ''' check we write variant attributes correctly (Aside from genotype)
        '''
        path = self.tmpdir / 'temp.bgen'
        sample_ids = ['a', 'b', 'c']
        with BgenWriter(path, n_samples=3, samples=sample_ids) as bfile:
            geno = np.array([[0.1, 0.8, 0.1], [0.5, 0.25, 0.25], [0.1, 0.2, 0.7]])
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,)
            bfile.add_variant('var2', 'rs2', 'chr1', 11, ['C', 'T'], geno / 2)
        
        with BgenReader(path, delay_parsing=True) as bfile:
            self.assertEqual(bfile.samples, sample_ids)

            # check that we've written all the attributes correctly
            var = next(bfile)
            self.assertEqual(var.varid, 'var1')
            self.assertEqual(var.rsid, 'rs1')
            self.assertEqual(var.chrom, 'chr1')
            self.assertEqual(var.pos, 10)
            self.assertEqual(var.alleles, ['A', 'C'])

            # and check the second variant
            var = next(bfile)
            self.assertEqual(var.varid, 'var2')
            self.assertEqual(var.rsid, 'rs2')
            self.assertEqual(var.chrom, 'chr1')
            self.assertEqual(var.pos, 11)
            self.assertEqual(var.alleles, ['C', 'T'])

    def test_writing_genotypes(self):
        ''' test BgenWriter
        '''
        path = self.tmpdir / 'temp.bgen'
        with BgenWriter(path, n_samples=3, samples=['a', 'b', 'c'],
                           compression=None, layout=2, metadata='1234') as bfile:
            geno = np.array([[0.1, 0.8, 0.1], 
                            [0.5, 0.25, 0.25], 
                            [float('nan'), float('nan'), float('nan')],
                            ])
            bit_depth = 16
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno, bit_depth=bit_depth)
            bfile.add_variant('var2', 'rs2', 'chr1', 11, ['C', 'T'], geno / 2, bit_depth=bit_depth)
        
        with BgenReader(path, delay_parsing=True) as bfile:
            self.assertEqual(bfile.samples, ['a', 'b', 'c'])

            # check the first variant works
            # check the genotype probabilities are correct, but last sample is all nan
            var = next(bfile)
            self.assertTrue(np.isnan(var.probabilities[2, :]).all())
            self.assertTrue(probs_close(geno[:, :-1], var.probabilities[:, :-1], bit_depth))
            
            # and check the second variant works
            var = next(bfile)
            # adjust the genotype probabilities for the second variant
            geno = geno / 2
            geno[:, -1] = 1 - geno[:, :-1].sum(axis=1)
            self.assertTrue(probs_close(geno[:, :-1], var.probabilities[:, :-1], bit_depth))

            # we hit the end after two variants
            with self.assertRaises(StopIteration):
                next(bfile)

    def test_writing_many_genotypes(self):
        ''' test BgenWriter with a few thousand samples
        '''
        path = self.tmpdir / 'temp.bgen'
        n_samples = 2000
        a = np.linspace(0, 0.3, n_samples)
        b = np.linspace(0.3, 0.6, n_samples)
        geno = np.vstack([a, b, 1 - (a + b)]).T
        with BgenWriter(path, n_samples=n_samples, layout=1, compression=None) as bfile:
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno)
        
        with BgenReader(path, delay_parsing=True) as bfile:
            var = next(bfile)
            self.assertTrue(probs_close(geno[:, :-1], var.probabilities[:, :-1], bit_depth=8))

    def test_compression_and_layouts(self):
        compressions = [None, 'zlib', 'zstd']
        layouts = [1, 2]
        geno = np.array([[0.1, 0.8, 0.1], 
                        [0.5, 0.25, 0.25], 
                        [float('nan'), float('nan'), float('nan')],
                        ])

        idx = 1
        for compression in compressions:
            for layout in layouts:
                if compression == 'zstd' and layout == 1:
                    continue
                path = path = self.tmpdir / f'temp_{idx}.bgen'
                idx += 1
                with BgenWriter(path, 3, samples=['a', 'b', 'c'],
                                compression=compression, layout=layout) as bfile:
                    bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno)

                with BgenReader(path, delay_parsing=True) as bfile:
                    self.assertEqual(bfile.header.compression, compression)
                    self.assertEqual(bfile.header.layout, layout)
                    for x in bfile:
                        probs = x.probabilities
                        self.assertTrue(probs_close(geno[:, :-1], probs[:, :-1], 8))

    def test_bit_depths(self):
        ''' check writing to different bit depths works
        '''
        geno = np.array([[0.1, 0.8, 0.1], 
                        [0.5, 0.25, 0.25], 
                        [float('nan'), float('nan'), float('nan')],
                        ])
        
        idx = 1
        for bit_depth in range(1, 33):
            path = path = self.tmpdir / f'temp_{idx}.bgen'
            idx += 1
            with BgenWriter(path, 3, samples=['a', 'b', 'c']) as bfile:
                bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                                bit_depth=bit_depth)

            with BgenReader(path, delay_parsing=True) as bfile:
                for x in bfile:
                    probs = x.probabilities
                    self.assertTrue(probs_close(geno[:, :-1], probs[:, :-1], bit_depth))
    
    def test_more_alleles(self):
        ''' check writing to different bit depths works
        '''
        path = self.tmpdir / 'temp.bgen'
        geno1 = np.array([[0.1, 0.8, 0.1],
                        [0.5, 0.25, 0.25],
                        [float('nan'), float('nan'), float('nan')],
                        ])
        geno2 = np.array([[0.1, 0.6, 0, 0, 0.1, 0.2],
                        [0.1, 0.2, 0.1, 0.2, 0.1, 0.3],
                        [float('nan'), float('nan'), float('nan'), float('nan'), float('nan'), float('nan')],
                        ])
        bit_depth = 8
        bfile = BgenWriter(path, 3, samples=['a', 'b', 'c'])
        bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno1,
                        bit_depth=bit_depth)
        bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C', 'T'], geno2,
                        bit_depth=bit_depth)
        bfile.close()

        bfile = BgenReader(path, delay_parsing=True)
        x = next(bfile)
        self.assertTrue(probs_close(geno1[:, :-1], x.probabilities[:, :-1], bit_depth))

        x = next(bfile)
        self.assertEqual(len(x.alleles), 3)
        self.assertTrue(probs_close(geno2[:, :-1], x.probabilities[:, :-1], bit_depth))

    def test_phased_data(self):
        '''checking writing phased data'''
        path = self.tmpdir / 'temp.bgen'
        bfile = BgenWriter(path, 3, samples=['a', 'b', 'c'])
        geno = np.array([[0.1, 0.9, 0.5, 0.5], 
                        [0.2, 0.8, 0.4, 0.6],
                        [float('nan'), float('nan'), float('nan'), float('nan')]])
        bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                            phased=1, bit_depth=8)
        bfile.close()

        bfile = BgenReader(path, delay_parsing=True)
        for x in bfile:
            probs = x.probabilities
            self.assertTrue(probs_close(geno[:, :-1], probs[:, :-1], bit_depth=8))
    
    def test_phased_data_many_samples(self):
        '''checking writing phased data with many samples.
        
        This also hits a fast path for parsing phased data'''
        path = self.tmpdir / 'temp.bgen'
        n_samples = 1000
        with BgenWriter(path, n_samples) as bfile:
            # construct a genotype array where the values
            a = np.linspace(0, 0.3, n_samples)
            b = np.linspace(0.7, 1, n_samples)
            geno = np.vstack([a, 1-a, b, 1-b]).T
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                                phased=1, bit_depth=8)

        time.sleep(1)

        with BgenReader(path, delay_parsing=True) as bfile:
            for x in bfile:
                probs = x.probabilities
                self.assertTrue(probs_close(geno[:, :-1], probs[:, :-1], bit_depth=8))
    
    def test_phased_data_different_sizes(self):
        '''checking writing phased data with a range of sample sizes'''
        for n_samples in range(1, 100):
            path = self.tmpdir / f'temp_{n_samples}.bgen'
            bfile = BgenWriter(path, n_samples)
            # construct a genotype array where the values
            a = np.linspace(0, 0.3, n_samples)
            b = np.linspace(0.7, 1, n_samples)
            geno = np.vstack([a, 1-a, b, 1-b]).T
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                                phased=1, bit_depth=8)
            bfile.close()

            bfile = BgenReader(path, delay_parsing=True)
            for x in bfile:
                probs = x.probabilities
                self.assertTrue(probs_close(geno[:, :-1], probs[:, :-1], bit_depth=8))
            bfile.close()
    
    def test_ploidy_unphased(self):
        ''' check we can write unphased variants with variable ploidy per sample
        '''
        path = self.tmpdir / 'temp.bgen'
        bfile = BgenWriter(path, 3, samples=['a', 'b', 'c'])
        ploidy = np.array([1, 2, 3], dtype=np.uint8)
        geno = np.array([[0.1, 0.9, float('nan'), float('nan')], 
                        [0.2, 0.4, 0.4, float('nan')],
                        [float('nan'), float('nan'), float('nan'), float('nan')]])
        bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                            ploidy=ploidy)
        bfile.close()

        bfile = BgenReader(path, delay_parsing=True)
        for x in bfile:
            probs = x.probabilities
            self.assertTrue(probs_close(geno, probs, bit_depth=8))
        
    def test_ploidy_phased(self):
        ''' check we can write phased variants with variable ploidy per sample
        '''
        path = self.tmpdir / 'temp.bgen'
        bfile = BgenWriter(path, 4, samples=['a', 'b', 'c', 'd'])
        ploidy = np.array([1, 2, 3, 3], dtype=np.uint8)
        geno = np.array([[0.1, 0.9, float('nan'), float('nan'), float('nan'), float('nan')], 
                         [0.2, 0.8, 0.5, 0.5, float('nan'), float('nan')],
                         [float('nan'), float('nan'), float('nan'), float('nan'), float('nan'), float('nan')],
                         [0.3, 0.7, 0.2, 0.8, 1, 0],
                         ])
        bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                            ploidy=ploidy, phased=1)
        bfile.close()

        bfile = BgenReader(path, delay_parsing=True)
        for x in bfile:
            probs = x.probabilities
            self.assertTrue(probs_close(geno, probs, bit_depth=8))
    
    def test_add_variant_direct(self):
        ''' test adding variant data directly from BgenVar
        '''
        path_1 = self.tmpdir / 'temp.bgen'
        bfile = BgenWriter(path_1, 4, samples=['a', 'b', 'c', 'd'])
        ploidy = np.array([1, 2, 3, 3], dtype=np.uint8)
        geno = np.array([[0.1, 0.9, float('nan'), float('nan'), float('nan'), float('nan')],
                         [0.2, 0.8, 0.5, 0.5, float('nan'), float('nan')],
                         [float('nan'), float('nan'), float('nan'),
                          float('nan'), float('nan'), float('nan')],
                         [0.3, 0.7, 0.2, 0.8, 1, 0],
                         ])
        bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                          ploidy=ploidy, phased=1)

        # and add another, slightly different, variant
        geno = np.array([[0.1, 0.7, float('nan'), float('nan'), 0.3, 0.4],
                         [0.2, 0.7, 0.5, 0.5, float('nan'), float('nan')],
                         [float('nan'), float('nan'), float('nan'),
                          float('nan'), float('nan'), float('nan')],
                         [0.3, 0.1, 0.2, 0.8, 0.5, 0],
                         ])
        bfile.add_variant('var2', 'rs2', 'chr2', 20, ['G', 'TT'], geno,
                          ploidy=ploidy, phased=1)
        bfile.close()

        path_2 = self.tmpdir / 'temp2.bgen'
        with BgenReader(path_1) as bfile:
            samples = bfile.samples
            with BgenWriter(path_2, len(samples), samples) as output:
                for var in bfile:
                    output.add_variant_direct(var)

        with BgenReader(path_1) as bfile_1, BgenReader(path_2) as bfile_2:
            for var1, var2 in zip(bfile_1, bfile_2):
                self.assertEqual(var1.rsid, var2.rsid)
                self.assertEqual(var1.chrom, var2.chrom)
                self.assertEqual(var1.pos, var2.pos)
                self.assertEqual(var1.alleles, var2.alleles)
                self.assertEqual(var1.fileoffset, var2.fileoffset)
                self.assertEqual(var1.next_variant_offset, var2.next_variant_offset)

                # check all the nan values match
                self.assertTrue((np.isnan(var1.probabilities) == np.isnan(var2.probabilities)).all())

                # check all the non-nan values match
                mask = np.isfinite(var1.probabilities)
                self.assertTrue((var1.probabilities[mask] == var2.probabilities[mask]).all())

    def test_multiple_read_writes(self):
        ''' check values pass correctly through multiple rounds of read/writes
        '''
        first_path = self.tmpdir / 'temp1.bgen'
        second_path = self.tmpdir / 'temp2.bgen'
        
        bit_depth = 9
        max_val = (2 ** bit_depth) - 1
        half = max_val // 2
        integer_values = [[0,                max_val,     0],
                          [half,             half + 1,    0],
                          [half + 1,         half,        0],
                          [(max_val - 1),    1,           0],
                         ]
        integer_values = np.array(integer_values)
        geno = integer_values / max_val
        
        with BgenWriter(first_path, n_samples=len(geno)) as bfile:
            bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno, bit_depth=bit_depth)
        
        with BgenWriter(second_path, n_samples=len(geno)) as out_bfile:
            with BgenReader(first_path, delay_parsing=True) as bfile:
                probs = next(bfile).probabilities
                as_integers = (probs * max_val).round()
                self.assertTrue(probs_close(geno, probs, bit_depth=bit_depth))
                self.assertTrue((integer_values == as_integers).all())
                out_bfile.add_variant(
                    'var1', 'rs1', 'chr1', 10, ['A', 'C'], probs, bit_depth=bit_depth)
        
        with BgenReader(second_path, delay_parsing=True) as bfile:
                probs = next(bfile).probabilities
                as_integers = (probs * max_val).round()
                self.assertTrue(probs_close(geno, probs, bit_depth=bit_depth))
                self.assertTrue((integer_values == as_integers).all())
    
    def test_all_possible_genotypes(self):
        ''' check all possible values in the range available to the bit depths
        '''
        max_samples = 10000000
        for bit_depth in range(1, 24):
            first_path = self.tmpdir / f'temp_{bit_depth}.v1.bgen'
            second_path = self.tmpdir / f'temp2_{bit_depth}.v2.bgen'
            
            max_val = (2 ** bit_depth) - 1
            increment = 1
            if (max_val / 2) > max_samples:
                increment = ((max_val + 1) / 2) / max_samples
            
            integer_values = np.arange(0, max_val + 1, increment)
            
            remainder = max_val - (integer_values[::2] + integer_values[::-2])
            
            integer_values = np.array([integer_values[::2], integer_values[::-2],
                                       remainder], dtype=np.uint32)
            integer_values = np.ascontiguousarray(integer_values.T)
            geno = integer_values / max_val
            
            # write first round
            with BgenWriter(first_path, n_samples=len(geno)) as bfile:
                bfile.add_variant('var1', 'rs1', 'chr1', 10, ['A', 'C'], geno,
                                  bit_depth=bit_depth)
            
            # check the first write
            with BgenWriter(second_path, n_samples=len(geno)) as out_bfile:
                with BgenReader(first_path, delay_parsing=True) as bfile:
                    probs = next(bfile).probabilities
                    as_integers = (probs * max_val).round()
                    self.assertTrue(probs_close(geno, probs, bit_depth=bit_depth))
                    self.assertTrue((integer_values == as_integers).all())
                    out_bfile.add_variant(
                        'var1', 'rs1', 'chr1', 10, ['A', 'C'], probs, bit_depth=bit_depth)
            
            # check the re-written bgen
            with BgenReader(second_path, delay_parsing=True) as bfile:
                    probs = next(bfile).probabilities
                    as_integers = (probs * max_val).round()
                    self.assertTrue(probs_close(geno, probs, bit_depth=bit_depth))
                    self.assertTrue((integer_values == as_integers).all())




from pathlib import Path
import unittest

import numpy as np

from bgen import BgenReader

from tests.utils import load_gen_data

class TestBgenReader(unittest.TestCase):
    ''' class to make sure BgenReader works correctly
    '''
    
    @classmethod
    def setUpClass(cls):
        cls.gen_data = load_gen_data()
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        self.folder = Path(__file__).parent /  "data"
    
    def test_context_handler_closed_bgen_samples(self):
        ''' no samples available from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile.samples) > 0)
        
        with self.assertRaises(ValueError):
            bfile.samples
    
    def test_context_handler_closed_bgen_varids(self):
        ''' no varids available from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile.varids()) > 0)
        
        with self.assertRaises(ValueError):
            bfile.varids()
    
    def test_context_handler_closed_bgen_rsids(self):
        ''' no rsids available from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile.rsids()) > 0)
        
        with self.assertRaises(ValueError):
            bfile.rsids()
    
    def test_context_handler_closed_bgen_positions(self):
        ''' no positions available from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile.positions()) > 0)
        
        with self.assertRaises(ValueError):
            bfile.positions()
    
    def test_context_handler_closed_bgen_length(self):
        ''' error raised if accessing length of exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile) > 0)
        
        with self.assertRaises(ValueError):
             len(bfile)
    
    def test_context_handler_closed_bgen_slice(self):
        ''' error raised if slicing variant from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile) > 0)
        
        with self.assertRaises(ValueError):
             var = bfile[0]
    
    def test_context_handler_closed_bgen_at_position(self):
        ''' error raised if getting variant at position from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile) > 0)
        
        with self.assertRaises(ValueError):
             var = bfile.at_position(100)
    
    def test_context_handler_closed_bgen_with_rsid(self):
        ''' error raised if getting variant with rsid from exited BgenReader
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            self.assertTrue(len(bfile) > 0)
        
        with self.assertRaises(ValueError):
             var = bfile.with_rsid('rs111')
    
    def test_context_handler_variant_data_not_loaded(self):
        ''' error raised if we try to access variant data after closing BgenFile
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            var = next(bfile)
        
        with self.assertRaises(ValueError):
            # cannot access data after file closed
            var.minor_allele_dosage
    
    def test_context_handler_variant_data_loaded(self):
        '''no error raised for variant from closed BgenReader, IF data is already loaded
        '''
        path = self.folder / 'example.16bits.zstd.bgen'
        with BgenReader(path) as bfile:
            var = next(bfile)
            var.minor_allele_dosage # load data while file still open
        
        # can access data after file closed, but only because the file was read 
        # previously while still open
        dose = var.minor_allele_dosage
        self.assertTrue(isinstance(dose, np.ndarray))
    
    def test_fetch(self):
        ''' can fetch variants within a genomic region
        '''
        chrom, start, stop = '01', 5000, 50000
        bfile = BgenReader(self.folder / 'example.16bits.bgen')
        self.assertTrue(bfile._check_for_index(str(self.folder / 'example.16bits.bgen')))
        
        self.assertTrue(list(bfile.fetch('02')) == [])
    
    def test_fetch_whole_chrom(self):
        ''' fetching just with chrom gives all variants on chromosome
        '''
        chrom, start, stop = '01', 5000, 50000
        bfile = BgenReader(self.folder / 'example.16bits.bgen')
        
        # test fetching a whole chromosome
        sortkey = lambda x: (x.chrom, x.pos)
        for x, y in zip(sorted(bfile.fetch(chrom), key=sortkey), sorted(self.gen_data, key=sortkey)):
            self.assertEqual(x.rsid, y.rsid)
            self.assertEqual(x.chrom, y.chrom)
            self.assertEqual(x.pos, y.pos)
    
    def test_fetch_after_position(self):
        ''' fetching variants with chrom and start gives all variants after pos
        '''
        chrom, start, stop = '01', 5000, 50000
        bfile = BgenReader(self.folder / 'example.16bits.bgen')
        
        sortkey = lambda x: (x.chrom, x.pos)
        gen_vars = [x for x in sorted(self.gen_data, key=sortkey) if start <= x.pos]
        for x, y in zip(sorted(bfile.fetch(chrom, start), key=sortkey), gen_vars):
            self.assertEqual(x.rsid, y.rsid)
            self.assertEqual(x.chrom, y.chrom)
            self.assertEqual(x.pos, y.pos)
    
    def test_fetch_in_region(self):
        ''' fetching variants with chrom, start, stop gives variants in region
        '''
        chrom, start, stop = '01', 5000, 50000
        bfile = BgenReader(self.folder / 'example.16bits.bgen')
        
        sortkey = lambda x: (x.chrom, x.pos)
        gen_vars = [x for x in sorted(self.gen_data, key=sortkey) if start <= x.pos <= stop]
        for x, y in zip(sorted(bfile.fetch(chrom, start, stop), key=sortkey), gen_vars):
            self.assertEqual(x.rsid, y.rsid)
            self.assertEqual(x.chrom, y.chrom)
            self.assertEqual(x.pos, y.pos)
        
        # check that we don't get any variants in a region without any
        self.assertEqual(list(bfile.fetch(chrom, start * 1000, stop * 1000)), [])




import glob
import io
import os
from pathlib import Path
from setuptools import setup
import subprocess
import sys
import platform

from distutils.core import Extension
from distutils.ccompiler import new_compiler
from Cython.Build import cythonize

EXTRA_COMPILE_ARGS = []
EXTRA_LINK_ARGS = []
if sys.platform == 'linux':
    EXTRA_COMPILE_ARGS += ['-std=c++11', '-I/usr/include', '-O2']
elif sys.platform == "darwin":
    EXTRA_COMPILE_ARGS += [
        "-stdlib=libc++",
        "-std=c++11",
        "-O2",
        "-I/Library/Developer/CommandLineTools/SDKs/MacOSX.sdk/usr/include/c++/v1",
        "-I/Library/Developer/CommandLineTools/SDKs/MacOSX.sdk/usr/include",
        ]
    EXTRA_LINK_ARGS += [
        "-L/Library/Developer/CommandLineTools/SDKs/MacOSX.sdk/usr/lib",
        ]
elif sys.platform == "win32":
    EXTRA_COMPILE_ARGS += ['/std:c++14', '/O2']

if platform.machine() == 'x86_64' and sys.platform != "darwin":
    EXTRA_COMPILE_ARGS += ['-mavx', '-mavx2']

def flatten(*lists):
    return [str(x) for sublist in lists for x in sublist]

def build_zlib():
    ''' compile zlib code to object files for linking
    
    Returns:
        list of paths to compiled object code
    '''
    cur_dir = Path.cwd()
    source_dir = cur_dir / 'src' / 'zlib-ng'
    build_dir = cur_dir / 'zlib_build'
    build_dir.mkdir(exist_ok=True)
    os.chdir(build_dir)
    
    cmd = ['cmake', '-S', source_dir, '-B', build_dir,
        '-DZLIB_COMPAT=ON',
        '-DZLIB_ENABLE_TESTS=OFF',
        '-DBUILD_SHARED_LIBS=OFF',
        '-DCMAKE_C_FLAGS="-fPIC"',
    ]
    subprocess.run(cmd)
    subprocess.run(['cmake', '--build', build_dir, '-v', '--config', 'Release'])
    os.chdir(cur_dir)
    
    objs = [str(build_dir / 'libz.a')]
    if sys.platform == 'win32':
        objs = [str(build_dir / 'Release' / 'zlibstatic.lib'),
                ]
    
    return str(build_dir), objs

def build_zstd():
    ''' compile zstd code to object files for linking with bgen c++ code
    
    This needs to be compiles independently of the bgen c++ code, as zstd is in
    c, so cannot be compiled directly with the bgen code. zstd is compiled to
    object files, and these are staticly linked in with the bgen code.
    
    I tried to link to a shared/dynamic zstd library, but couldn't specify a
    location that would be found by the compiled bgen library after relocating
    files into to a python package.
    
    Returns:
        list of paths to compiled object code
    '''
    print('building zstd')
    folder = Path('src/zstd/lib')
    include_dirs = ['src/zstd/lib/', 'src/zstd/lib/common']
    sources = flatten(
        (folder / 'common').glob('*.c'),
        (folder / 'compress').glob('*.c'),
        (folder / 'decompress').glob('*.c'),
        (folder / 'dictBuilder').glob('*.c'),
        (folder / 'deprecated').glob('*.c'),
        (folder / 'legacy').glob('*.c'),  # TODO: drop some legacy versions
    )
    extra_compile_args = ['-std=gnu11', '-fPIC', '-O2']
    
    # newer zstd versions have an asm file, which needs to be to compiled and
    # used as a library for full zstd compilation.
    cc = new_compiler()
    # the unix compiler needs to allow files with .S extension
    cc.src_extensions += ['.S']
    compiled = cc.compile(sources=flatten((folder / 'decompress').glob('*.S')),)
    
    if len(compiled) > 0:
        extra_compile_args += [f'-L{" ".join(compiled)}']
    
    return cc.compile(sources, include_dirs=include_dirs,
        extra_preargs=extra_compile_args) + compiled

zlib_dir, zlib = build_zlib()
zstd = build_zstd()

extensions = [
    Extension('bgen.reader',
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
        sources=['src/bgen/reader.pyx',
            'src/reader.cpp',
            'src/genotypes.cpp',
            'src/header.cpp',
            'src/samples.cpp',
            'src/utils.cpp',
            'src/variant.cpp'],
        extra_objects=zstd + zlib,
        include_dirs=['src', 'src/zstd/lib', zlib_dir],
        language='c++'),
    Extension('bgen.writer',
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
        sources=['src/bgen/writer.pyx',
            'src/writer.cpp',
            'src/genotypes.cpp',
            'src/utils.cpp',
            ],
        extra_objects=zstd + zlib,
        include_dirs=['src', 'src/zstd/lib', zlib_dir],
        language='c++'),
    ]

setup(
    package_dir={'': 'src'},
    ext_modules=cythonize(extensions),
    test_loader='unittest:TestLoader',
    )



# cython: language_level=3, boundscheck=False, emit_linenums=True

import logging
from pathlib import Path
import sqlite3
import sys
import time

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t

from cython.operator cimport dereference as deref

import numpy as np

cdef extern from "<iostream>" namespace "std":
    cdef cppclass ostream:
        pass

cdef extern from "<iostream>" namespace "std::ios_base":
    cdef cppclass open_mode:
        pass
    cdef open_mode binary

cdef extern from "<fstream>" namespace "std":
    cdef cppclass ofstream(ostream):
        ofstream() except +
        ofstream(const string&) except +
        ofstream(const string&, open_mode) except +

cdef extern from 'genotypes.h' namespace 'bgen':
    uint32_t get_max_probs(int &max_ploidy, int &n_alleles, bool &phased)

cdef extern from 'writer.h' namespace 'bgen':
    cdef cppclass CppBgenWriter:
        # declare class constructor and methods
        CppBgenWriter(string &path, uint32_t n_samples, string &free_data, 
                   uint32_t compression, uint32_t layout, vector[string] &samples) except +
        uint64_t write_variant_header(string &varid, string &rsid, string &chrom, 
                        uint32_t &pos, vector[string] &alleles, uint32_t _n_samples) except +
        uint64_t write_variant_direct(vector[uint8_t] & data) except +
        uint64_t add_genotype_data(uint16_t n_alleles,
                         double *genotypes, uint32_t geno_len, uint8_t ploidy,
                         bool phased, uint8_t bit_depth) except +
        uint64_t add_genotype_data(uint16_t n_alleles,
                         double *genotypes, uint32_t geno_len, uint8_t *ploidy,
                         uint32_t min_ploidy, uint32_t max_ploidy,
                         bool phased, uint8_t bit_depth) except +

class Indexer:
    ''' class to automatically index bgen files as they are being constructed
    '''
    def __init__(self, bgen_path):
        index_path = Path(str(bgen_path) + '.bgi')
        if index_path.exists():
            index_path.unlink()
        self.conn = sqlite3.connect(index_path)
        self.cur = self.conn.cursor()
        self.create_tables()
    
    def create_tables(self):
        query = '''CREATE TABLE Metadata (
                    filename TEXT NOT NULL, 
                    file_size INT NOT NULL, 
                    last_write_time INT NOT NULL, 
                    first_1000_bytes BLOB NOT NULL, 
                    index_creation_time INT NOT NULL)'''
        self.cur.execute(query)
        query = '''CREATE TABLE Variant (
                    chromosome TEXT NOT NULL,
                    position INT NOT NULL,
                    rsid TEXT NOT NULL,
                    number_of_alleles INT NOT NULL,
                    allele1 TEXT NOT NULL,
                    allele2 TEXT NULL,
                    file_start_position INT NOT NULL,
                    size_in_bytes INT NOT NULL,
                PRIMARY KEY (chromosome, position, rsid, allele1, allele2, file_start_position))
                WITHOUT ROWID'''
        self.cur.execute(query)
        
        # index the Variant table
        self.cur.execute('CREATE INDEX chrom_index on Variant(chromosome)')
        self.cur.execute('CREATE INDEX pos_index on Variant(position)')
        self.cur.execute('CREATE INDEX rsid_index on Variant(rsid)')
    
    def add_variant(self, chrom, pos, rsid, alleles, offset, size):
        query = '''INSERT INTO Variant VALUES (?, ?, ?, ?, ?, ?, ?, ?)'''
        params = (chrom, pos, rsid, len(alleles), alleles[0], alleles[1], offset, size)
        self.cur.execute(query, params)
    
    def close(self):
        self.conn.commit()
        if sys.platform == 'win32':
            time.sleep(0.01)
        self.cur.close()

cdef class BgenWriter:
    ''' class to open bgen files from disk, and access variant data within
    '''
    cdef CppBgenWriter * thisptr
    cdef string path
    cdef bool is_open
    cdef object indexer
    def __cinit__(self, path, uint32_t n_samples, samples=[], compression='zstd',
                  layout=2, metadata=None):
        if isinstance(path, Path):
            path = str(path)
        
        if compression not in [None, 'zstd', 'zlib']:
            raise ValueError(f'compression type {compression} not one of zlib or zstd')
        
        cdef uint32_t compress_flag=0
        if compression == 'zlib':
            compress_flag = 1
        elif compression == 'zstd':
            compress_flag = 2
        
        # re-define variables into cpp objects
        cdef string _metadata = metadata.encode('utf8') if metadata is not None else b''
        cdef vector[string] _samples = [x.encode('utf8') for x in samples]

        self.path = path.encode('utf8')
        
        logging.debug(f'opening CppBgenWriter from {self.path.decode("utf")}')
        self.thisptr = new CppBgenWriter(self.path, n_samples, _metadata, compress_flag, layout, _samples)
        self.is_open = True
        self.indexer = Indexer(path)
    
    def __dealloc__(self):
        self.close()
    
    def __repr__(self):
        return f'BgenFile("{self.path.decode("utf8")}")'
    
    def add_variant(self, varid, rsid, chrom, uint32_t pos, alleles, 
                    genotypes, ploidy=2, bool phased=False,
                    uint8_t bit_depth=8):
        ''' add a variant to the bgen file on disk

        Args:
            varid: variant ID
            rsid: reference SNP ID
            chrom: chromosome the variant is on
            pos: nucleotide position of the variant
            alleles: list of allele strings
            genotypes: numpy array of genotype proabilities, ordered as per the
                bgen samples.
            ploidy: integer for constant ploidy, or numpy array of ploidy values per 
                sample, in same order as genotypes
            phased: whether the genotypes are for phased data or not
            bit_depth: interger from 1-32 (inclusive) for how many bits to store
                each genotype in.
        '''

        # re-define variables into cpp objects
        cdef string _varid = varid.encode('utf8')
        cdef string _rsid = rsid.encode('utf8')
        cdef string _chrom = chrom.encode('utf8')
        cdef vector[string] _alleles = [x.encode('utf8') for x in alleles]
        cdef uint32_t n_samples = len(genotypes)

        if not self.is_open:
            raise ValueError("bgen file is closed")
        var_offset = self.thisptr.write_variant_header(_varid, _rsid, _chrom, pos, _alleles, n_samples)

        # determine ploidy levels
        cdef uint32_t ploidy_n=0
        cdef uint8_t[:] ploidy_arr = np.array([], dtype=np.uint8)
        if isinstance(ploidy, int):
            ploidy_n = ploidy
        elif isinstance(ploidy, np.ndarray):
            ploidy_arr = ploidy
        else:
            raise ValueError('ploidy must be either integer, or numpy array of integers')
        
        # convert numpy array to C contiguous for storing values on disk. numpy
        # arrays default to C contiguous, so most won't need conversion, but
        # some can be fortran order, e.g. if transposed
        if not isinstance(genotypes[0][0], np.float64):
            genotypes = genotypes.astype(np.float64)
        
        cdef double[:, :] geno_c
        if genotypes.flags['C_CONTIGUOUS']:
            geno_c = genotypes
        else:
            geno_c = np.ascontiguousarray(genotypes)
        
        cdef geno_len = genotypes.shape[0] * genotypes.shape[1]
        if len(ploidy_arr) == 0:
            end_offset = self.thisptr.add_genotype_data(_alleles.size(), &geno_c[0, 0],
                                           geno_len, ploidy_n, phased, bit_depth)
        else:
            end_offset = self.thisptr.add_genotype_data(_alleles.size(), &geno_c[0, 0],
                                           geno_len, &ploidy_arr[0], min(ploidy), 
                                           max(ploidy), phased, bit_depth)
        
        self.indexer.add_variant(chrom, int(pos), rsid, alleles, var_offset, 
                                 end_offset - var_offset)

    def add_variant_direct(self, variant):
        ''' insert a BgenVar directly into the bgen file
        '''
        chrom = variant.chrom
        pos = int(variant.pos)
        rsid = variant.rsid
        alleles = variant.alleles
        cdef vector[uint8_t] data = variant.copy_data()
        var_offset = self.thisptr.write_variant_direct(data)
        end_offset = var_offset + len(data)

        self.indexer.add_variant(chrom, pos, rsid, alleles, var_offset, end_offset - var_offset)

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False
    
    def close(self):
        if self.is_open:
            del self.thisptr
        if self.indexer is not None:
            self.indexer.close()
        self.is_open = False
        self.indexer = None
        if sys.platform == 'win32':
            time.sleep(0.01)
